#include <stdio.h>
#include <ctype.h>
#include <vector>
#include <algorithm>
#include <errno.h>
#include <string.h>
#include "linear.h"
#include "eval.h"

#define Malloc(type, n) (type *)malloc((n) * sizeof(type))

// Prototypes of evaluation functions
double precision(const size_t, const double *dec_values, const int *ty);
double recall(const size_t, const double *dec_values, const int *ty);
double fscore(const size_t, const double *dec_values, const int *ty);
double bac(const size_t, const double *dec_values, const int *ty);
double auc(const size_t, const double *dec_values, const int *ty);
double accuracy(const size_t, const double *dec_values, const int *ty);
double ap(const size_t, const double *dec_values, const int *ty);

// Evaluation function pointer
// You can assign this pointer to any above prototype
double (*evaluation_function)(const size_t, const double *, const int *) = accuracy;

bool quiet;
double positive_weight = 1;

// Evaluation functions below are the same for both LIBSVM and LIBLINEAR

double precision(const size_t size, const double *dec_values, const int *ty)
{
	size_t i;
	int tp, fp;
	double precision;

	tp = fp = 0;

	for (i = 0; i < size; ++i)
		if (dec_values[i] >= 0)
		{
			if (ty[i] == 1)
				++tp;
			else
				++fp;
		}

	if (tp + fp == 0)
	{
		if (!quiet)
			fprintf(stderr, "warning: No postive predict label.\n");
		precision = 0;
	}
	else
		precision = tp / (double)(tp + fp);
	if (!quiet)
		if (!quiet)
			printf("Precision = %g%% (%d/%d)\n", 100.0 * precision, tp, tp + fp);

	return precision;
}

double recall(const size_t size, const double *dec_values, const int *ty)
{
	size_t i;
	int tp, fn; // true_positive and false_negative
	double recall;

	tp = fn = 0;

	for (i = 0; i < size; ++i)
		if (ty[i] == 1)
		{ // true label is 1
			if (dec_values[i] >= 0)
				++tp; // predict label is 1
			else
				++fn; // predict label is -1
		}

	if (tp + fn == 0)
	{
		if (!quiet)
			fprintf(stderr, "warning: No postive true label.\n");
		recall = 0;
	}
	else
		recall = tp / (double)(tp + fn);
	// print result in case of invocation in prediction
	if (!quiet)
		printf("Recall = %g%% (%d/%d)\n", 100.0 * recall, tp, tp + fn);

	return recall; // return the evaluation value
}

double fscore(const size_t size, const double *dec_values, const int *ty)
{
	size_t i;
	int tp, fp, fn;
	double precision, recall;
	double fscore;

	tp = fp = fn = 0;

	for (i = 0; i < size; ++i)
		if (dec_values[i] >= 0 && ty[i] == 1)
			++tp;
		else if (dec_values[i] >= 0 && ty[i] == -1)
			++fp;
		else if (dec_values[i] < 0 && ty[i] == 1)
			++fn;

	if (tp + fp == 0)
	{
		// fprintf(stderr, "warning: No positive predict label.\n");
		precision = 0;
	}
	else
		precision = tp / (double)(tp + fp);
	if (tp + fn == 0)
	{
		// fprintf(stderr, "warning: No positive true label.\n");
		recall = 0;
	}
	else
		recall = tp / (double)(tp + fn);

	if (!quiet)
		printf("tp=%d fp=%d fn=%d p=%f r=%f\n", tp, fp, fn, precision, recall);
	if (precision + recall == 0)
	{
		// fprintf(stderr, "warning: precision + recall = 0.\n");
		fscore = 0;
	}
	else
		fscore = 2 * precision * recall / (precision + recall);

	if (!quiet)
		printf("F-score = %g\n", fscore);

	return fscore;
}

double bac(const size_t size, const double *dec_values, const int *ty)
{
	size_t i;
	int tp, fp, fn, tn;
	double specificity, recall;
	double bac;

	tp = fp = fn = tn = 0;

	for (i = 0; i < size; ++i)
		if (dec_values[i] >= 0 && ty[i] == 1)
			++tp;
		else if (dec_values[i] >= 0 && ty[i] == -1)
			++fp;
		else if (dec_values[i] < 0 && ty[i] == 1)
			++fn;
		else
			++tn;

	if (tn + fp == 0)
	{
		if (!quiet)
			fprintf(stderr, "warning: No negative true label.\n");
		specificity = 0;
	}
	else
		specificity = tn / (double)(tn + fp);
	if (tp + fn == 0)
	{
		if (!quiet)
			fprintf(stderr, "warning: No positive true label.\n");
		recall = 0;
	}
	else
		recall = tp / (double)(tp + fn);

	bac = (specificity + recall) / 2;
	if (!quiet)
		printf("BAC = %g\n", bac);

	return bac;
}

// only for auc
class Comp
{
	const double *dec_val;

public:
	Comp(const double *ptr) : dec_val(ptr) {}
	bool operator()(int i, int j) const
	{
		return dec_val[i] > dec_val[j];
	}
};

double auc(const size_t size, const double *dec_values, const int *ty)
{
	double roc = 0;
	size_t i;
	std::vector<size_t> indices(size);

	for (i = 0; i < size; ++i)
		indices[i] = i;

	std::sort(indices.begin(), indices.end(), Comp(&dec_values[0]));

	int tp = 0, fp = 0;
	for (i = 0; i < size; i++)
	{
		if (ty[indices[i]] == 1)
			tp++;
		else if (ty[indices[i]] == -1)
		{
			roc += tp;
			fp++;
		}
	}

	if (tp == 0 || fp == 0)
	{
		fprintf(stderr, "warning: Too few postive true labels or negative true labels\n");
		roc = 0;
	}
	else
		roc = roc / tp / fp;

	if (!quiet)
		printf("AUC = %g\n", roc);

	return roc;
}

double accuracy(const size_t total, const double *dec_values, const int *ty)
{
	int correct = 0;
	size_t i;

	for (i = 0; i < total; ++i)
		if (ty[i] == (dec_values[i] >= 0 ? 1 : -1))
			++correct;

	if (!quiet)
		printf("Accuracy = %g%% (%d/%d)\n",
			   (double)correct / (int)total * 100, correct, (int)total);

	return (double)correct / (int)total;
}

double ap(const size_t size, const double *dec_values, const int *ty)
{
	size_t i;
	std::vector<size_t> indices(size);

	for (i = 0; i < size; ++i)
		indices[i] = i;
	std::sort(indices.begin(), indices.end(), Comp(&dec_values[0]));

	int p = 0, tp = 0;
	double prev_recall = 0, area = 0;

	for (i = 0; i < size; ++i)
		p += (ty[i] == 1);

	if (p == 0)
	{
		fprintf(stderr, "warning: Too few postive labels\n");
		return 0;
	}

	for (i = 0; i < size; ++i)
	{
		tp += (ty[indices[i]] == 1);

		if (i + 1 < size && dec_values[indices[i]] == dec_values[indices[i + 1]])
			continue;

		double recall = (double)tp / p;
		double precision = (double)tp / (double)(i + 1);

		area += precision * (recall - prev_recall);
		prev_recall = recall;
	}

	if (!quiet)
		printf("AP = %g\n", area);
	return area;
}

int check_binary_beforeCV(const problem *prob, int *ty)
{
	int label[3];
	int nr_class = 0;
	int is_binary = 0;

	for (int i = 0; i < prob->l; i++)
	{
		int this_label = (int)prob->y[i];
		int j;
		for (j = 0; j < nr_class; j++)
			if (this_label == label[j])
				break;
		if (j == nr_class)
		{
			if (nr_class > 2)
				break;
			label[nr_class] = this_label;
			++nr_class;
		}
		if (this_label == 1)
			ty[i] = 1;
		else if (this_label == -1)
			ty[i] = -1;
	}
	if (nr_class == 2)
	{
		if ((label[0] == 1 && label[1] == -1) ||
			(label[0] == -1 && label[1] == 1))
			is_binary = 1;
		else
		{
			fprintf(stderr, "ERROR: to use other evaluation criteria, labels must be +1/-1\n");
			exit(1);
		}
	}

	return is_binary;
}

static void exit_input_error(int line_num)
{
	fprintf(stderr, "Wrong input format at line %d\n", line_num);
	exit(1);
}

static char *line = NULL;
static int max_line_len;

static char *readline(FILE *input)
{
	int len;

	if (fgets(line, max_line_len, input) == NULL)
		return NULL;

	while (strrchr(line, '\n') == NULL)
	{
		max_line_len *= 2;
		line = (char *)realloc(line, max_line_len);
		len = (int)strlen(line);
		if (fgets(line + len, max_line_len - len, input) == NULL)
			break;
	}
	return line;
}

void binary_class_predict(FILE *input, FILE *output, const model *model_)
{
	int total = 0;
	int *labels;
	int max_nr_attr = 64;
	struct feature_node *x = Malloc(struct feature_node, max_nr_attr);
	int max_nr_data = 64;
	double *dec_values = Malloc(double, max_nr_data);
	int *true_labels = Malloc(int, max_nr_data);
	int n;
	if (model_->bias >= 1)
		n = get_nr_feature(model_) + 1;
	else
		n = get_nr_feature(model_);

	int nr_class = get_nr_class(model_);
	labels = Malloc(int, nr_class);
	get_labels(model_, labels);

	if (nr_class == 2)
		for (int j = 0; j < 2; j++)
			if (labels[j] != 1 && labels[j] != -1)
			{
				fprintf(stderr, "ERROR: to use other evaluation criteria, labels must be +1/-1\n");
				exit(1);
			}

	max_line_len = 1024;
	line = (char *)malloc(max_line_len * sizeof(char));
	while (readline(input) != NULL)
	{
		int i = 0;
		double target_label, predict_label;
		char *idx, *val, *label, *endptr;
		int inst_max_index = 0; // strtol gives 0 if wrong format

		label = strtok(line, " \t\n");
		if (label == NULL) // empty line
			exit_input_error(total + 1);

		target_label = strtod(label, &endptr);
		if (endptr == label || *endptr != '\0')
			exit_input_error(total + 1);

		while (1)
		{
			if (i >= max_nr_attr - 2) // need one more for index = -1
			{
				max_nr_attr *= 2;
				x = (struct feature_node *)realloc(x, max_nr_attr * sizeof(struct feature_node));
			}

			idx = strtok(NULL, ":");
			val = strtok(NULL, " \t");

			if (val == NULL)
				break;
			errno = 0;
			x[i].index = (int)strtol(idx, &endptr, 10);
			if (endptr == idx || errno != 0 || *endptr != '\0' || x[i].index <= inst_max_index)
				exit_input_error(total + 1);
			else
				inst_max_index = x[i].index;

			errno = 0;
			x[i].value = strtod(val, &endptr);
			if (endptr == val || errno != 0 || (*endptr != '\0' && !isspace(*endptr)))
				exit_input_error(total + 1);

			++i;
		}

		if (model_->bias >= 0)
		{
			x[i].index = n;
			x[i].value = model_->bias;
			++i;
		}
		x[i].index = -1;

		predict_label = predict(model_, x);
		fprintf(output, "%.17g\n", predict_label);

		double dec_value;
		predict_values(model_, x, &dec_value);

		true_labels[total] = (target_label > 0) ? 1 : -1;
		dec_values[total] = dec_value;

		total++;
		if (total >= max_nr_data)
		{
			max_nr_data *= 2;
			dec_values = (double *)realloc(dec_values, max_nr_data * sizeof(double));
			true_labels = (int *)realloc(true_labels, max_nr_data * sizeof(int));
		}
	}

	evaluation_function(total, dec_values, true_labels);

	free(labels);
	free(x);
	free(dec_values);
	free(true_labels);
}

eval_func get_evaluation_function()
{
	return evaluation_function;
}

double weighted_error(const size_t total, const double *dec_values, const int *ty)
{
	// use negative error because CV maximizes metric
	double err = 0;

	for (size_t i = 0; i < total; ++i)
	{
		if (ty[i] > 0)
		{
			// false negative
			err -= positive_weight * (dec_values[i] < 0);
		}
		else
		{
			// false positive
			err -= (dec_values[i] >= 0);
		}
	}

	if (!quiet)
		printf("weighted error = %g (%.1f/%d)\n",
			   err / total, err, int(total));

	return err / total;
}
