% csne is cost-sensitive-no-empty.
% Quick and dirty implementation, reuses cost-sensitive weights.
function csne(train_path, test_path)
    rand('seed', 1);
    start_time = clock;

    if (nargin ~= 2)
        fprintf('Incorrect number of input parameters\n');
        return;
    end

    data_name = extractBetween(train_path, 'data/', '/train');
    split = extractBetween(train_path, 'train', '.svm');

    % w has dimensions number of features * number of labels
    % b has dimensions number of labels * 1 or 1 * 1
    model_path = sprintf('models/%s%s_cost-sensitive', ...
        strrep(char(data_name), '/', '_'), char(split));
    vars = load(model_path);
    w = vars.w;
    b = vars.b;
    options = vars.options; % options is dummy

    [train_data, test_data, dummy] = ...
        prepare_data(train_path, test_path, options);

    % (w' * x)' has dimensions number of instances * number of labels
    % b' has dimensions 1 * number of labels
    % p has dimensions number of instances * number of labels
    p = ((w' * test_data.x)' + b');

    [val, idx] = max(p, [], 2);
    no_p = find(val < 0)';
    p(sub2ind(size(p), no_p, idx(no_p)')) = 1;

    method = 'cost-sensitive-no-empty';

    rloss = ranking_loss(2*test_data.y - 1, p);
    [macro, micro, instance] = fscore(2*test_data.y - 1, 2*(p > 0) - 1);
    fprintf('%s%s %s macro = %.6f micro = %.6f instance = %.6f rloss = %.6f elapsed = %.1fs\n', ...
        char(data_name), char(split), method, macro, micro, instance, rloss, etime(clock, start_time))
    
% options will be parsed and stripped of -B,
% with bias padded as a feature for x.
% number of features includes the bias.
% x is a sparse matrix.
% x has dimensions number of features * number of instances
% y is a dense 0/1 matrix.
% y has dimensions number of instances * number of labels
function [train_data, test_data, options] = prepare_data(train_path, test_path, options)
    tmp = strfind(options, '-R');
    if isempty(tmp) == 0
        fprintf('-R is not supported\n');
        exit;
    end

    bias = 0;
    options_split = strsplit(options, ' ');
    tmp = strfind(options_split, '-B');
    for i = 1:size(tmp, 2),
        if tmp{i} == 1,
            bias = str2num(options_split{i+1});
            options = [strjoin(options_split(1,1:i-1)) ' ' strjoin(options_split(1,i+2:end))];
            break;
        end
    end
    
    [y, x, map] = libsvmread_ml(train_path);
    [ty, tx, tmap] = libsvmread_ml(test_path);

    x = x';
    tx = tx';

    % eliminate instances that have duplicated label
    y = full(y);
    ty = full(ty);

    [labels, ord, tord] = intersect(map, tmap);
    y = y(:, ord);
    ty = ty(:, tord);

    if numel(labels) ~= numel(map) || numel(labels) ~= numel(tmap)
        fprintf('WARNING: train/test labels do not match\n');
    end

    t0all = clock;

    % fill x
    probk = size(x, 1);
    testk = size(tx, 1);
    k = max(probk, testk);
    x = [x; zeros(k-probk, size(y, 1))];
    tx = [tx; zeros(k-testk, size(ty, 1))];

    if bias > 0
        x=[x; ones(1, size(x,2))*bias];
        tx=[tx; ones(1, size(tx,2))*bias];
    end

    train_data = struct;
    train_data.x = x;
    train_data.y = y;
    train_data.map = labels;
    test_data = struct;
    test_data.x = tx;
    test_data.y = ty;
    test_data.map = labels;
