#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mex.h"

#ifdef MX_API_VER
#if MX_API_VER < 0x07030000
typedef int mwIndex;
#endif
#endif

void exit_with_help()
{
	mexPrintf(
	"Usage: libsvmwrite('filename', label_matrix, instance_matrix, map_vector);\n"
	);
}

static void fake_answer(int nlhs, mxArray *plhs[])
{
	int i;
	for(i=0;i<nlhs;i++)
		plhs[i] = mxCreateDoubleMatrix(0, 0, mxREAL);
}

void libsvmwrite(const char *filename, const mxArray *label_mat, const mxArray *instance_mat, const mxArray *map_vec)
{
	FILE *fp = fopen(filename,"w");
	mwIndex *l_ir, *l_jc, *ir, *jc, k, low, high;
	size_t i, l, label_mat_row_num;
	double *samples, *labels, *map;
	mxArray *label_mat_col, *instance_mat_col; // sparse matrices in column format

	if(fp ==NULL)
	{
		mexPrintf("can't open output file %s\n",filename);
		return;
	}

	// transpose label matrix
	{
		mxArray *prhs[1], *plhs[1];
		prhs[0] = mxDuplicateArray(label_mat);
		if(mexCallMATLAB(1, plhs, 1, prhs, "transpose"))
		{
			mexPrintf("Error: cannot transpose label matrix\n");
			return;
		}
		label_mat_col = plhs[0];
		mxDestroyArray(prhs[0]);
	}
	
	// transpose instance matrix
	{
		mxArray *prhs[1], *plhs[1];
		prhs[0] = mxDuplicateArray(instance_mat);
		if(mexCallMATLAB(1, plhs, 1, prhs, "transpose"))
		{
			mexPrintf("Error: cannot transpose instance matrix\n");
			return;
		}
		instance_mat_col = plhs[0];
		mxDestroyArray(prhs[0]);
	}

	// the number of instance
	l = mxGetN(instance_mat_col);
	label_mat_row_num = mxGetM(label_mat);

	if(label_mat_row_num!=l)
	{
		mexPrintf("# instances of label and instance matrices do not match.\n");
		return;
	}

	// each column is one instance
	labels = mxGetPr(label_mat_col);
	l_ir = mxGetIr(label_mat_col);
	l_jc = mxGetJc(label_mat_col);
	samples = mxGetPr(instance_mat_col);
	ir = mxGetIr(instance_mat_col);
	jc = mxGetJc(instance_mat_col);

	map = mxGetPr(map_vec);

	for(i=0;i<l;i++)
	{
		low = l_jc[i], high = l_jc[i+1];
		for(k=low;k<high-1;k++)
			fprintf(fp,"%.17g,", map[l_ir[k]]);
		if (low <= high-1)
			fprintf(fp,"%.17g", map[l_ir[high-1]]);
		
		low = jc[i], high = jc[i+1];
		for(k=low;k<high;k++)
			fprintf(fp," %lu:%g", (size_t)ir[k]+1, samples[k]);

		fprintf(fp,"\n");
	}

	fclose(fp);
	return;
}

void mexFunction( int nlhs, mxArray *plhs[],
		int nrhs, const mxArray *prhs[] )
{
	if(nlhs > 0)
	{
		exit_with_help();
		fake_answer(nlhs, plhs);
		return;
	}

	// Transform the input Matrix to libsvm format
	if(nrhs == 4)
	{
		char filename[256];
		if(!mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]) || !mxIsDouble(prhs[3]))
		{
			mexPrintf("Error: label matrix, instance matrix, and label mappings must be double\n");
			return;
		}

		mxGetString(prhs[0], filename, mxGetN(prhs[0])+1);

		if(mxIsSparse(prhs[1]) && mxIsSparse(prhs[2]))
			libsvmwrite(filename, prhs[1], prhs[2], prhs[3]);
		else
		{
			mexPrintf("Label and instance matrices must be sparse\n");
			return;
		}
	}
	else
	{
		exit_with_help();
		return;
	}
}
