function unified_run(train_path, test_path, options, method)
    rand('seed', 1);
    start_time = clock;

    if (nargin ~= 4)
        fprintf('Incorrect number of input parameters\n');
        return;
    end

    [train_data, test_data, prepared_options] = ...
        prepare_data(train_path, test_path, options);

    data_name = extractBetween(train_path, 'data/', '/train');
    split = extractBetween(train_path, 'train', '.svm');

    % w has dimensions number of features * number of labels
    % b has dimensions number of labels * 1 or 1 * 1
    if startsWith(method, 'resume-')
        method = method(8:size(method,2));
        model_path = sprintf('models/%s%s_%s', ...
            strrep(char(data_name), '/', '_'), char(split), method);
        vars = load(model_path);
        if ~strcmp('', options) && ~strcmp(vars.options, options)
            error('Cannot resume with inconsistent options\n\tparameter: %s\n\tloaded: %s', ...
                options, vars.options)
        end
        w = vars.w;
        b = vars.b;
    else
        switch method
        case {'unrealistic', 'one-vs-rest', 'one-vs-rest-no-empty'}
            [w, b] = one_vs_rest(train_data, prepared_options);
        case 'one-vs-rest-C'
            [w, b] = one_vs_rest_C(train_data, prepared_options);
        case 'thresholding'
            [w, b] = thresholding(train_data, prepared_options);
        case 'cost-sensitive'
            [w, b] = cost_sensitive(train_data, prepared_options);
        case 'cost-sensitive-simple'
            [w, b] = cost_sensitive_simple(train_data, prepared_options);
        case 'cost-sensitive-simple-micro'
            [w, b] = cost_sensitive_simple_micro(train_data, prepared_options);
        otherwise
            error('Unknown method %s', method)
        end
        model_path = sprintf('models/%s%s_%s', ...
            strrep(char(data_name), '/', '_'), char(split), method);
        save(model_path, 'w', 'b', 'options');
    end

    % (w' * x)' has dimensions number of instances * number of labels
    % b' has dimensions 1 * number of labels
    % p has dimensions number of instances * number of labels
    p = ((w' * test_data.x)' + b');

    switch method
    case 'unrealistic'
        [dummy, ord] = sort(p, 2, 'descend');
        take = sum(test_data.y, 2);
        p = zeros(size(p)) - 1; % this breaks ranking loss
        for i = 1:size(ord, 1)
            idx = ord(i, 1:take(i));
            p(i, idx) = take(i):-1:1;
        end
    case 'one-vs-rest-no-empty'
        [val, idx] = max(p, [], 2);
        no_p = find(val < 0)';
        p(sub2ind(size(p), no_p, idx(no_p)')) = 1;
    end

    rloss = ranking_loss(2*test_data.y - 1, p);
    [macro, micro, instance] = fscore(2*test_data.y - 1, 2*(p > 0) - 1);
    fprintf('%s%s %s macro = %.6f micro = %.6f instance = %.6f rloss = %.6f elapsed = %.1fs\n', ...
        char(data_name), char(split), method, macro, micro, instance, rloss, etime(clock, start_time))
    
% options will be parsed and stripped of -B,
% with bias padded as a feature for x.
% number of features includes the bias.
% x is a sparse matrix.
% x has dimensions number of features * number of instances
% y is a dense 0/1 matrix.
% y has dimensions number of instances * number of labels
function [train_data, test_data, options] = prepare_data(train_path, test_path, options)
    tmp = strfind(options, '-R');
    if isempty(tmp) == 0
        fprintf('-R is not supported\n');
        exit;
    end

    bias = 0;
    options_split = strsplit(options, ' ');
    tmp = strfind(options_split, '-B');
    for i = 1:size(tmp, 2),
        if tmp{i} == 1,
            bias = str2num(options_split{i+1});
            options = [strjoin(options_split(1,1:i-1)) ' ' strjoin(options_split(1,i+2:end))];
            break;
        end
    end
    
    [y, x, map] = libsvmread_ml(train_path);
    [ty, tx, tmap] = libsvmread_ml(test_path);

    x = x';
    tx = tx';

    % eliminate instances that have duplicated label
    y = full(y);
    ty = full(ty);

    [labels, ord, tord] = intersect(map, tmap);
    y = y(:, ord);
    ty = ty(:, tord);

    if numel(labels) ~= numel(map) || numel(labels) ~= numel(tmap)
        fprintf('WARNING: train/test labels do not match\n');
    end

    t0all = clock;

    % fill x
    probk = size(x, 1);
    testk = size(tx, 1);
    k = max(probk, testk);
    x = [x; zeros(k-probk, size(y, 1))];
    tx = [tx; zeros(k-testk, size(ty, 1))];

    if bias > 0
        x=[x; ones(1, size(x,2))*bias];
        tx=[tx; ones(1, size(tx,2))*bias];
    end

    train_data = struct;
    train_data.x = x;
    train_data.y = y;
    train_data.map = labels;
    test_data = struct;
    test_data.x = tx;
    test_data.y = ty;
    test_data.map = labels;
